
#include "quakedef.h"



typedef enum {
	pt_static, pt_grav, pt_slowgrav, pt_fire, pt_explode, pt_explode2, pt_blob, pt_blob2
} zptype_t;

// !!! if this is changed, it must be changed in d_ifacea.h too !!!
typedef struct zparticle_s
{
// driver-usable fields
	vec3_t		org;
	float		color;
// drivers never touch the following fields

	struct zparticle_s	* next;

	vec3_t		vel;
	float		ramp;
	float		die;
	float		born;
	zptype_t		type;
	int			fxtype;
	int			tex_num;
	int			drawtype;
	float		scale;
	float		alpha;

	float		angle;
	float		v_angle;
} zparticle_t;


int LoadParticleFile(char *part);
void R_DrawSpriteParticle(zparticle_t *p, float age, float ttlive, float distance, float grav, float frametime);
void R_DrawDartParticle(zparticle_t *p, float age, float ttlive, float distance, float grav, float frametime);
void R_DrawGlowParticle(zparticle_t *p, float age, float ttlive, float distance, float grav, float frametime);
void ZR_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count, int type);

static int			zr_numparticles = 0;

zparticle_t	*zactive_particles, *zfree_particles;

zparticle_t	*zparticles = 0;


float	rotater[4];
int		over_dim = 1;

int		zsmoke_tex, zfire_tex, zradar_tex, zexplo_tex, zdebris_tex, zspark_tex, zblood_tex;


extern int		ramp1[8];
extern int		ramp2[8];
extern int		ramp3[8];


extern  int     b_smoke1_tex;
extern int     b_smoke2_tex;
extern  int     b_smoke3_tex;
extern  int     b_smoke4_tex;
extern int builtin_particle_tex;
extern int		smoke1_tex;
extern int		fire_tex;
extern int      blood_tex;
extern int		particle_tex;

#define 	NEWZPARTICLE			if (!zfree_particles)\
					return;\
				p = zfree_particles;\
				zfree_particles = p->next;\
				p->next = zactive_particles;\
				zactive_particles = p;


/*=============
R_InitParticles
=============*/
void ZR_InitParticles (void)
{
	int i;


	
	zr_numparticles = 2048;


	//zparticles = (zparticle_t *) Hunk_AllocName ((zr_numparticles + 1) * sizeof(zparticle_t), "particles");
	
	zparticles = malloc((zr_numparticles + 1)*sizeof(zparticle_t));

	if (!zparticles)
		Sys_Error("No free memory for particle engine 4\n");


	memset(zparticles, 0, zr_numparticles * sizeof(zparticle_t));




	// load textures
	// setup rotation
	for (i=4; i>=0; i--)
		rotater[i] = i/4.0f * M_PI * 2;
	
	zsmoke_tex = LoadParticleFile("zr_smoke1");


	LoadParticleFile("zr_smoke2");
	LoadParticleFile("zr_smoke3");
	LoadParticleFile("zr_smoke4");

	if (!zsmoke_tex)
	{
		if (!smoke1_tex)
			zsmoke_tex = b_smoke1_tex;			
		else
			zsmoke_tex = smoke1_tex;
	}


	zfire_tex = LoadParticleFile("zr_fire1");
	LoadParticleFile("zr_fire2");
	LoadParticleFile("zr_fire3");
	LoadParticleFile("zr_fire4");
	LoadParticleFile("zr_fire5");
	LoadParticleFile("zr_fire6");

	if (!zfire_tex)
	{
		zfire_tex = fire_tex;
	}

	zradar_tex = LoadParticleFile("zr_radar");

	if (!zradar_tex)
		zradar_tex = particle_tex;

	zexplo_tex = LoadParticleFile("zr_explo1");
	i = LoadParticleFile("zr_explo2");
	i = LoadParticleFile("zr_explo3");
	i = LoadParticleFile("zr_explo4");

	if (!zexplo_tex)
	{
		zexplo_tex = smoke1_tex;
	}

	zdebris_tex = LoadParticleFile("zr_debris1");
	i = LoadParticleFile("zr_debris2");
	i = LoadParticleFile("zr_debris3");
	i = LoadParticleFile("zr_debris4");

	if (!zdebris_tex)
	{
		zdebris_tex = particle_tex;
	}

	zblood_tex = LoadParticleFile("zr_splat1");
	i = LoadParticleFile("zr_splat2");
	i = LoadParticleFile("zr_splat3");
	i = LoadParticleFile("zr_splat4");
	
	if (!zblood_tex)
	{
		zblood_tex = blood_tex;
	}

	zspark_tex = LoadParticleFile("zr_spark1");

	if (!zspark_tex)
	{
		zspark_tex = particle_tex;
	}

}

void R_CreateParticle(int count,
					  int fxtype,
					  int drawtype,
					  int tex_num,
					  float color,
					  zptype_t type,
					  float born,
					  float die,
					  float angle,
					  float v_angle,
					  float alpha,
					  float scale,
					  vec3_t org,
					  int org_rand,
					  int vel,
					  int vel_z,
					  int power)
{
	int i, j;
	zparticle_t *p;

	for(i=0; i<count;i++)
	{
		if (!zfree_particles)
			return;
		p = zfree_particles;
		zfree_particles = p->next;
		p->next = zactive_particles;
		zactive_particles = p;
		
		p->fxtype = fxtype;
		p->drawtype = drawtype;
		p->tex_num = tex_num;
		p->color = color;
		p->type = type;
		p->born = born;
		p->die = die;
		p->angle = angle;
		p->v_angle = v_angle;
		p->alpha = alpha;
		p->scale = scale;

		if (power == 0) power = 1;

		//origin
		if (org_rand == 0)
		{
			VectorCopy(org, p->org);
		}
		else
		{
			for (j=0 ; j<3 ; j++)
				p->org[j] = org[j] + (rand()%(org_rand*2))-org_rand;
		}

		// velocity
		if (vel_z != 0 && vel != 0)
		{
			for (j=0 ; j<2 ; j++)
				p->vel[j] = ((rand()%(vel*2))-vel) * power;
			p->vel[2] = (rand()%vel_z) + power;
		}
		else if (vel != 0)
		{
			for (j=0 ; j<3 ; j++)
				p->vel[j] = ((rand()%(vel*2))-vel) * power;
		}
		else if (vel_z != 0)
		{
			p->vel[2] = (rand()%vel_z) + power;
			p->vel[0] = 0.0f;
			p->vel[1] = 0.0f;
		}
		else
		{
			p->vel[0] = 0.0f;
			p->vel[1] = 0.0f;
			p->vel[2] = 0.0f;
		}

	}
}

void oldZR_ClearParticles (void)
{
	int		i;
	
	

	zfree_particles = &zparticles[0];
	zactive_particles = NULL;


	for (i=0 ;i<zr_numparticles ; i++)
		zparticles[i].next = &zparticles[i+1];

	zparticles[zr_numparticles-1].next = NULL;
}


void ZR_ClearParticles (void)
{
	int		i;
	void * p;

	
	size_t size;

	p = (void *)zparticles;

	zfree_particles			= &zparticles[0];
	zactive_particles		= NULL;
	
	size = zr_numparticles * sizeof(zparticle_t);

	memset(p, 0, size);
	
	for( i = 0 ; i < zr_numparticles ; ++i )
	{
//		zparticles[i].next	= (zparticle_t * )((int )(zparticles) + (int)(sizeof(zparticle_t) * (i + 1)));
		zparticles[i].next = &zparticles[i+1];

	}
	
	zparticles[zr_numparticles - 1].next	= NULL;

}



void ZR_DrawParticles (void)
{
	zparticle_t		*p, *kill;
	float			grav;
	int				i, j, bob, lastcase;
	float			time2, time3;
	float			time1;
	float			dvel;
	float			frametime;
	float			ttlive;
	float			alpha;
	float			age;
	float			rander;
	
	vec3_t			up, right;
	float			scale;

	vec3_t	v;
	float		radius;
	float		distance, cutoff;
	float		intensity;
	extern cvar_t sv_gravity;
	
	if (gl_fogenable.value)
		glDisable(GL_FOG);
	glDepthMask (FALSE);
//	glDisable (GL_TEXTURE_2D);
//	glShadeModel (GL_SMOOTH);
	//glEnable (GL_BLEND);
	//glBlendFunc (GL_ONE, GL_ONE);
	glPushMatrix();			

	VectorScale (vup, 1, up);
	VectorScale (vright, 1, right);

	frametime = cl.time - cl.oldtime;
	time3 = frametime * 15;
	time2 = frametime * 10; // 15;
	time1 = frametime * 5;
	grav = frametime * sv_gravity.value * 0.05;
	dvel = 4*frametime;
	
	for ( ;; ) 
	{
		kill = zactive_particles;
		if (kill && kill->die < cl.time)
		{
			zactive_particles = kill->next;
			kill->next = zfree_particles;
			zfree_particles = kill;
			continue;
		}
		break;
	}

	lastcase = PFX_GLOW;

	for (p=zactive_particles ; p ; p=p->next)
	{
		for ( ;; )
		{
			kill = p->next;
			if (kill && kill->die < cl.time)
			{
				p->next = kill->next;
				kill->next = zfree_particles;
				zfree_particles = kill;
				continue;
			}
			break;
		}

		// time to live per particle
		ttlive = p->die - cl.time;
		//if (ttlive > 5 * r_particleduration.value)
		if (ttlive > 5 )
			p->die = cl.time + 0.01f;
		age = cl.time - p->born;
		VectorSubtract(p->org, r_origin, v);
		distance = Length(v);

		bob = p->drawtype;
		switch(bob)
		{
			case PFX_HUD:
				if (lastcase != PFX_HUD)
				{
					glDisable (GL_DEPTH_TEST);
					glEnable (GL_TEXTURE_2D);
					//glEnable (GL_BLEND);
					//???//glBlendFunc (GL_SRC_ALPHA, GL_ONE);
					glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
					lastcase = PFX_SPRITE;
				}
				R_DrawSpriteParticle(p, age, ttlive, distance, grav, frametime);
				break;

			case PFX_SPRITE:
				if (lastcase != PFX_SPRITE)
				{
					//if (lastcase == PFX_HUD)
					//glEnable (GL_DEPTH_TEST);
					glEnable (GL_TEXTURE_2D);
					//glEnable (GL_BLEND);
					glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
					lastcase = PFX_SPRITE;
				}
				glEnable (GL_DEPTH_TEST);
				R_DrawSpriteParticle(p, age, ttlive, distance, grav, frametime);
				break;

			case PFX_DART:
				if (lastcase != PFX_DART)
				{
					glEnable (GL_DEPTH_TEST);
					glDisable (GL_TEXTURE_2D);
					//glEnable (GL_BLEND);
					glBlendFunc (GL_ONE, GL_ONE);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
					lastcase = p->drawtype;
				}
				R_DrawDartParticle(p, age, ttlive, distance, grav, frametime);
				break;
			
			case PFX_GLOW:
			case PFX_TWINKLE:
			default:
				if (lastcase != PFX_GLOW || lastcase != PFX_TWINKLE)
				{
					//if (lastcase == PFX_HUD)
						glEnable (GL_DEPTH_TEST);
					glDisable (GL_TEXTURE_2D);
					//glEnable (GL_BLEND);
					glBlendFunc (GL_ONE, GL_ONE);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
					lastcase = p->drawtype;
				}
				R_DrawGlowParticle(p, age, ttlive, distance, grav, frametime);
				break;
		}
		p->org[0] += p->vel[0]*frametime;
		p->org[1] += p->vel[1]*frametime;
		p->org[2] += p->vel[2]*frametime;

		p->angle += (p->v_angle * frametime);
		p->angle = fmod(p->angle, 360.0f);
	}

	glPopMatrix();
	glColor3f (1,1,1);
//	glDisable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_DEPTH_TEST);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask (TRUE);
	if (gl_fogenable.value)
		glEnable(GL_FOG);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	

}

void R_DrawSpriteParticle(zparticle_t *p, float age, float ttlive, float distance, float grav, float frametime)
{
	float		size, red, green, blue, scale, limit, angle;
	vec3_t		up, right, point;
	int			i,j, bob;

	bob = p->fxtype;
	switch(bob)
	{
		default:
		case TE_TRAIL:
			//p->scale = 20 + (3*r_particleduration.value - ttlive) * 7;
			p->scale = 20 + (3 - ttlive) * 7;
			p->alpha = ttlive * 0.1;
			p->v_angle = p->v_angle * 0.93;
			red = green = blue = over_dim;
			break;

		case TE_THRUST:
			//p->scale = 5 + (6*r_particleduration.value - ttlive) * 3;
			p->scale = 5 + (6 - ttlive) * 3;
			p->alpha = ttlive * 0.1;
			red = green = blue = (1 - (cl.time - p->born));
			p->v_angle *= 0.95f;
			VectorScale(p->vel, 1 - (0.85 * frametime), p->vel);
			p->vel[2] += 2 * frametime;
			break;

		case TE_EXPLOSION:
			p->scale = 20 + ((5 - ttlive) * 40);
			p->alpha = ttlive * 0.25;
			red = green = blue = (1 - (cl.time - p->born));
			p->v_angle *= 0.95f;
			VectorScale(p->vel, 0.97f, p->vel);
			break;

		case TE_FIRE:
			//p->scale = 5 + (3*r_particleduration.value - ttlive) * 5;
			p->scale = 5 + (3 - ttlive) * 5;
			p->alpha = 1 - age * 2;
			p->vel[2] = p->vel[2] + 15;
			red = green = blue = over_dim;
			break;

		case TE_BLOODTRAIL:
			p->scale = 5 + (1 - ttlive) * 15;
			p->alpha = ttlive;
			red = over_dim * 0.5;
			green = blue = 0;
			break;
		
		case TE_RADARBLIP:
			p->scale = 32;
			p->alpha = ttlive * 0.1f;
			red = green = blue = over_dim;
			break;

		case TE_GUNSHOT:
			p->alpha = ttlive;
			red = green = blue = over_dim;
			p->vel[2] -= grav * 8;
			break;

		case TE_BLOOD:
			p->scale = 2 + (0.25 - ttlive) * 32;
			p->alpha = ttlive * 4;
			p->v_angle = p->v_angle * 0.93;
			red = green = blue = over_dim;
			break;

		case TE_TELEPORT:
			p->scale = 20 + (0.5 - ttlive) * 60;
			p->alpha = ttlive * 2;
			red = green = blue = over_dim;
			break;
	}

	VectorCopy(vright, right);
	VectorCopy(vup, up);

	// fudge size of nearby particles and scale by world
	size = p->scale ;
	limit = size * 5 + 1;
	if (distance < limit && bob != TE_FIRE)
		size = size * (distance/limit);

	// draw particle
	glColor4f(red, green, blue, p->alpha);
	glBindTexture (GL_TEXTURE_2D, p->tex_num);
	glBegin (GL_QUADS);

	for (i=4; i>=0; i--)
	{
		if (i == 1)
			glTexCoord2f (0, 0);
		else if (i == 2)
			glTexCoord2f (0, 1);
		else if (i == 3)
			glTexCoord2f (1, 1);
		else if (i == 4)
			glTexCoord2f (1, 0);
		else
			glTexCoord2f (1, 0);

		angle = rotater[i] + p->angle;
		for (j=0; j<3; j++)
			point[j] = p->org[j] + vright[j]*cos(angle)*size + vup[j]*sin(angle)*size;
		glVertex3fv(point);
	}
	glEnd ();
}

void R_DrawDartParticle(zparticle_t *p, float age, float ttlive, float distance, float grav, float frametime)
{
	int			i, j, bob;
	float		radius;
	float		cutoff;
	float		intensity;
	float		rander;
	float		red, green, blue;
	vec3_t		v, offset;
	vec3_t		forward, right, up;
	
	bob = p->fxtype;
	switch(bob)
	{
		case TE_SUPERSPIKE:
		case TE_SPIKE:
			radius = 1;
			intensity = ttlive * 4;
			red = intensity * 0.8f;
			green = intensity * 0.4f;
			blue = 0.1;
			cutoff = 1;
			p->vel[2] -= grav * 16;
			break;
	}

	glColor3f(red * over_dim, green * over_dim, blue * over_dim);
	
	if (radius <= 0.0f) radius = 1.0f;

	//radius *= worldscale.value;
	//cutoff *= worldscale.value;

	VectorScale(p->vel, -0.0625, offset);
	AngleVectors (offset, forward, right, up);
	VectorAdd(offset, p->org, offset);

	if (distance > cutoff)
	{
		glBegin(GL_TRIANGLE_FAN);
		for (i=0 ; i<3 ; i++)
			v[i] = p->org[i] - vpn[i] * radius;
		glVertex3fv(v);
		glColor3f(0.0f, 0.0f, 0.0f);
		for (i=16; i>=0; i--)
		{
			float a = i/16.0f * M_PI * 2;
			for (j=0; j<3; j++)
				v[j] = offset[j] + right[j]*cos(a)*radius + up[j]*sin(a)*radius;
			glVertex3fv(v);
		}
		glEnd();
	}
}


void R_DrawGlowParticle(zparticle_t *p, float age, float ttlive, float distance, float grav, float frametime)
{
	int			i, j, bob;
	float		radius;
	float		cutoff;
	float		intensity;
	float		rander;
	float		red, green, blue;
	vec3_t		v;
	
	bob = p->fxtype;
	switch(bob)
	{
		case TE_WIZSPIKE:
			radius = 5 + ttlive * 10;
			intensity = ttlive;
			red = 0.4f * intensity;
			green = 0.7f* intensity;
			blue = 0.1;
			cutoff = radius;
			break;

		case TE_KNIGHTSPIKE:
			radius = 5 + ttlive * 10;
			intensity = ttlive;
			red = intensity * 0.8f;
			green = intensity * 0.4f;
			blue = 0.1;
			cutoff = radius;
			break;

		case TE_SPARK:
			radius = 16;
			intensity = ttlive * 4;
			red = intensity * 0.8f;
			green = intensity * 0.4f;
			blue = 0.1;
			cutoff = radius;
			break;

		case TE_TRAIL:
//			radius = 10 + (2*r_particleduration.value - ttlive) * 5;
			radius = 10 + (2 - ttlive) * 5;
			intensity = ttlive * 0.1;
			red = green = blue = 0.2f * intensity;
			cutoff = radius * 5;
			break;

		case TE_FREE:
			radius = 1;
			glColor3ubv ((byte *)&d_8to24table[(int)p->color]);
			p->vel[2] -= grav * 32;
			cutoff = radius * 5;
			break;

		case TE_GUNSHOT:
			radius = 8;
			intensity = ttlive * 2;
			red = green = blue = intensity * 0.25f;
			cutoff = radius * 5;
			break;

		case TE_GLOWPUFF:
			radius = 12;
			intensity = ttlive * 2;
			red = intensity * 0.8f;
			green = intensity * 0.4f;
			blue = intensity * 0.1f;
			cutoff = radius * 5;
			break;

		case TE_SMALLPUFF:
//			radius = 8 + (4 * r_particleduration.value - ttlive) * 4;
			radius = 8 + (4  - ttlive) * 4;

			intensity = ttlive * 0.1;
			red = green = blue = intensity * 0.2f;
			p->vel[2] -= grav * 0.5;
			cutoff = radius * 5;
			break;
		
		case TE_SUPERSPIKE:
			radius = 8;
			intensity = ttlive * 2;
			red = green = blue = intensity * 0.25f;
			cutoff = radius * 5;
			break;

		case TE_EXPLOSION:
			radius = 100 + (255 - ttlive * 255);
			//radius = 30 + ttlive * 2048;
			intensity = ttlive;
			red = intensity * 0.8f;
			green = intensity * 0.4f;
			blue = intensity * 0.1f;
			cutoff = radius;
			break;

		case TE_TAREXPLOSION:
			radius = 30 - (ttlive * 6);
			intensity = ttlive * 0.2;
			red = intensity * 0.8f;
			green = intensity * 0.1f;
			blue = intensity * 0.6f;
			cutoff = radius * 5;
			break;

		case TE_LAVASPLASH:
			radius = 30 + ((1 - ttlive) * 40);
			intensity = 1 - age;
			red = intensity * 0.8f;
			green = intensity * 0.4f;
			blue = 0.1f;
			p->vel[2] -= grav * 8;
			cutoff = radius * 5;
			break;

		case TE_TELEPORT:
			radius = 10 - (ttlive * 0.1);
			intensity = ttlive * 3;
			red = green = blue = intensity * 1.0f;
			cutoff = radius * 5;
			break;

		case TE_EXPLOSION2:
			radius = 30 - (ttlive * 6);
			intensity = 1 - (age * 0.2); //ttlive * 0.2;
			red = intensity * 0.8f;
			green = intensity * 0.4f;
			blue = 0.1f;
			cutoff = radius * 5;
			break;

		case TE_BRIGHTFIELD:
			radius =  10;
			intensity = ttlive * 5;
			red = green = blue = intensity * 0.9f;
			cutoff = radius * 5;
			break;

		case TE_SMALLGLOW:
			radius =  20;
			red = 0.7f;
			green = 0.55f;
			blue = 0.4f;
			cutoff = radius * 5;
			break;

		case TE_BIGGLOW:
			radius =  50;
			red = 0.7f;
			green = 0.55f;
			blue = 0.4f;
			cutoff = radius * 5;
			break;

		case TE_BUBBLEGLOW:
			radius = 5;
			red = green = blue = 0.8f;
			cutoff = radius;
			break;
	
		case TE_SPARKGLOW:
			radius =  20;
			red = 0.7f;
			green = 0.55f;
			blue = 0.4f;
			cutoff = radius * 5;
			break;

		case TE_BOLTGLOW:
			radius =  20;
			red = 0.02f;
			green = 0.0f;
			blue = 0.1f;
			cutoff = radius;
			break;

		case TE_SHOCKWAVE:
			radius = 50 + (1 - ttlive) * 255;
			intensity = ttlive;
			red = intensity * 0.7f;
			green = intensity * 0.55f;
			blue = 0.4f;
			cutoff = radius;
			break;

		case TE_BIGSHOCKWAVE:
			radius = 100 + (255 - ttlive * 2047);
			glColor3f(0.35f, 0.24f, 0.2f);
			red = 0.35f;
			green = 0.24f;
			blue = 0.2f;
			cutoff = radius;
			break;

		case TE_SPARKWAVE:
			radius = 20 + (1 - ttlive) * 127;
			intensity = ttlive;
			red = intensity * 0.8f;
			green = intensity * 0.4f;
			blue = 0.1f;
			cutoff = radius;
			break;

		case TE_RAIN:
			radius =  2;
			intensity = (2-ttlive)*0.5;
			red = intensity * 0.2f;
			green = intensity * 0.2f;
			blue = 0.4f;
			p->vel[2] -= grav * 8;
			p->vel[1] = p->vel[1] + ((rand()%4)-2);
			p->vel[0] = p->vel[0] + ((rand()%4)-2);
			cutoff = radius * 5;
			break;
	
		case TE_SNOW:
			radius =  3;
			intensity = (4-ttlive)*0.5;
			red = intensity * 0.9f;
			green = intensity * 0.9f;
			blue = 0.9f;
			p->vel[2] -= grav * 2;
			p->vel[1] = p->vel[1] + ((rand()%16)-8);
			p->vel[0] = p->vel[0] + ((rand()%16)-8);
			cutoff = radius * 5;
			break;

		case TE_GOLDGLOW:
			radius =  35;
			red = 0.9f;
			green = 0.4f;
			blue = 0.1f;
			cutoff = radius;
			break;

		case TE_BLUEGLOW:
			radius =  35;
			red = 0.0f;
			green = 0.0f;
			blue = 0.5f;
			cutoff = radius;
			break;

		case TE_TORCHGLOW:
			radius = 30;
			intensity = 0.5 - ( (2000 - distance) * 0.00025 );
			red = intensity * 0.8f;
			green = intensity * 0.4f;
			blue = 0.1f;
			cutoff = radius;
			break;

		default:
			break;
	}

	if (bob != TE_FREE)
		glColor3f(red * over_dim, green * over_dim, blue * over_dim);
	
	if (radius <= 0.0f) radius = 1.0f;

//	radius *= worldscale.value;
//	cutoff *= worldscale.value;

	if (distance > cutoff)
	{
		glBegin(GL_TRIANGLE_FAN);
		for (i=0 ; i<3 ; i++)
			v[i] = p->org[i] - vpn[i] * radius;
		glVertex3fv(v);
		glColor3f(0.0f, 0.0f, 0.0f);
		for (i=16; i>=0; i--)
		{
			float a = i/16.0f * M_PI * 2;
			if (p->drawtype == PFX_TWINKLE)
			{
				rander = radius + radius * ((rand ()&0x7fff) / ((float)0x7fff));
				for (j=0; j<3; j++)
				 //v[j] =	p->org[j] + vright[j]*cos(a)*radius + vup[j]*sin(a)*radius;
				 v[j] =	p->org[j] + vright[j]*cos(a)* rander + vup[j]*sin(a)*rander;
			}
			else
			{
				for (j=0; j<3; j++)
					v[j] = p->org[j] + vright[j]*cos(a)*radius + vup[j]*sin(a)*radius;
			}
			glVertex3fv(v);
		}
		glEnd();
	}
}


void GunShot(vec3_t pos)
{
	ZR_RunParticleEffect (pos, vec3_origin, 0, 5, TE_GUNSHOT);
}

void ZR_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count, int type)
{
	count = count;
	if (type == TE_SUPERSPIKE)
		R_CreateParticle(count, TE_SUPERSPIKE, PFX_DART, 0, (color&~7) + (rand()&7), pt_slowgrav, cl.time, cl.time + 0.25, 0, 0, 1, 1, org, 0, 511, 255, 0);
	else if (type == TE_SPIKE)
		R_CreateParticle(count, TE_SPIKE, PFX_DART, 0, (color&~7) + (rand()&7), pt_slowgrav, cl.time, cl.time + 0.25, 0, 0, 1, 1, org, 0, 255, 127, 0);
	else if (type == TE_GUNSHOT)
	{
		R_CreateParticle(count, TE_GUNSHOT, PFX_SPRITE, zdebris_tex + (rand()%4), 95 + (rand()&15), pt_slowgrav, cl.time,
				cl.time + 1, rand()%7, (rand()%127 - 63), 1, 2 + (rand()%2), org, 0, 128, 200, 0);
		R_CreateParticle(2, TE_TRAIL, PFX_SPRITE, zsmoke_tex + (rand()%4), (color&~7) + (rand()&7), pt_slowgrav, cl.time,
			cl.time + 3, rand()%7, (rand()%31 - 15), 1, 1, org, 0, 15, 20, 0);
		//R_CreateParticle(2, TE_SPIKE, PFX_DART, 0, (color&~7) + (rand()&7), pt_slowgrav, cl.time, cl.time + 0.25, 0, 0, 1, 1, org, 0, 255, 127, 0);
	}
	else if (type == TE_SMALLPUFF)
	{
		R_CreateParticle(2, TE_TRAIL, PFX_SPRITE, zsmoke_tex + (rand()%4), (color&~7) + (rand()&7), pt_slowgrav, cl.time,
			cl.time + 3, rand()%7, (rand()%31 - 15), 1, 1, org, 0, 15, 40, 0);
	}
	else if (type == TE_GLOWPUFF)
	{
		R_CreateParticle(count, TE_GLOWPUFF, PFX_GLOW, 0, 107  + (rand()&3), pt_slowgrav, cl.time,
				cl.time + 0.1*(rand()%5), 0, 0, 0, 0, org, 0, 8, 20, 32);
	}
	else if (type == TE_BLOOD)
	{
		R_CreateParticle(count, TE_BLOOD, PFX_SPRITE, zblood_tex + (rand()%4), (color&~7) + (rand()&7), pt_slowgrav, cl.time,
			cl.time + 0.25, rand()%7, (rand()%31 - 15), 1, 1, org, 0, 63, 0, 0);
	}
	else
		R_CreateParticle(count, type, PFX_GLOW, 0, (color&~7) + (rand()&7), pt_slowgrav, cl.time,
			cl.time + 0.1*(rand()%5), 0, 0, 1, 1, org, 0, 50, 0, 0);
}

void ZR_ParticleExplosion (vec3_t org)
{
	int			i, j, k, count;
	zparticle_t	*p;
	vec3_t		dir;
	float		vel;

	count = 2;

	count = count * 1;

	for (k=0 ; k<count ; k++)
	{
		NEWZPARTICLE;

		p->die = cl.time + 1;
		p->born = cl.time;
		p->color = ramp1[0];
		p->fxtype = TE_EXPLOSION;
		p->drawtype = PFX_GLOW;
		p->type = pt_explode;

		for (j=0 ; j<3 ; j++)
		{
			p->org[j] = org[j] + (((rand()%32)-16));
			p->vel[j] = (rand()%128)-64;
		}
	}

	for (i=-count ; i<count ; i++)
		for (j=-count ; j<count ; j++)
			for (k=0 ; k<1 ; k++)
			{
				NEWZPARTICLE;

				p->die = cl.time + 2 + (rand()&2);
				p->born = cl.time;
				p->color = ramp1[0];
				p->type = pt_slowgrav;
				p->fxtype = TE_EXPLOSION;
				p->drawtype = PFX_SPRITE;
				p->tex_num = zexplo_tex;
				p->angle = rand()%7;
				p->v_angle = (rand()%40)-20;

				dir[0] = j*32 + (rand()&31);
				dir[1] = i*32 + (rand()&31);
				dir[2] = 512;
	
				p->org[0] = org[0] + (rand()&7);
				p->org[1] = org[1] + (rand()&7);
				p->org[2] = org[2] + (rand()&7);
	
				VectorNormalize (dir);						
				vel = 300 + (rand()&127);
				VectorScale (dir, vel, p->vel);
				if (i & 1)
					p->type = pt_explode;
				else
					p->type = pt_explode2;
			}

	for (i=-count ; i<count ; i++)
		for (j=-count ; j<count ; j++)
			for (k=0 ; k<1 ; k++)
			{
				NEWZPARTICLE;
		
				p->die = cl.time + 2;
				p->born = cl.time;
				p->color = ramp1[0];
				p->type = pt_slowgrav;
				p->fxtype = TE_EXPLOSION;
				p->drawtype = PFX_SPRITE;
				p->tex_num = zexplo_tex;
				p->angle = rand()%7;
				p->v_angle = (rand()%40)-20;
				
				dir[0] = j*32 + (rand()&15);
				dir[1] = i*32 + (rand()&15);
				dir[2] = (rand()&20);
	
				p->org[0] = org[0] + (rand()&7);
				p->org[1] = org[1] + (rand()&7);
				p->org[2] = org[2] + (rand()&7);
	
				VectorNormalize (dir);						
				vel = 300 + (rand()&127);
				VectorScale (dir, vel, p->vel);

				if (i & 1)
					p->type = pt_explode;
				else
					p->type = pt_explode2;
			}
}

void ZZR_RocketTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	zparticle_t	*p;
	int			dec;
	static int	tracercount;

	int fh1 = true;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);
	if (type < 128)
		dec = 10;
	else
	{
		dec = 1;
		type -= 128;
	}
	while (len > 0)
	{
		len -= dec;

	NEWZPARTICLE;		

		VectorCopy (vec3_origin, p->vel);

		switch (type)
		{
			case 0:	// rocket trail
				R_CreateParticle(1, TE_THRUST, PFX_SPRITE, zexplo_tex + (rand()%4), ramp3[(int)p->ramp], pt_fire, cl.time,
							 	 cl.time + 3 , rand()%7, (rand()%30)-15, 1, 1, start, 2, 25, 0, 0);
				break;

			case 1:	// smoke 
				R_CreateParticle(1, TE_TRAIL, PFX_SPRITE, zsmoke_tex + (rand()%4), (rand()&3) + 2, pt_fire, cl.time,
							 	 cl.time + 3 , rand()%7, (rand()%6)-3, 1, 1, start, 0, 10, 0, 0);
				break;


			case 2:	// blood
				R_CreateParticle(1, TE_BLOODTRAIL, PFX_SPRITE, zsmoke_tex + (rand()%4), 67 + (rand()&3), pt_grav, cl.time,
							 	 cl.time + 1, rand()%7, (rand()%6)-3, 1, 1, start, 3, 5, 0, 0);

				len -= 50;
				break;

			case 3:
			case 5:	// tracer
				p->type = pt_static;
				if (type == 3 && fh1)
				{
					p->die = cl.time + 0.5;
					p->born = cl.time;
					p->fxtype = TE_WIZSPIKE;
					p->drawtype = PFX_GLOW;
					p->color = 230 + ((tracercount&4)<<1);
					for (j=0; j<3 ; j++)
						p->vel[j] = (end[j] - start[j]) * 20;
					VectorCopy (start, p->org);
				}
				else if (type == 3 && !fh1)
				{
					p->die = cl.time + 1;
					p->born = cl.time;
					p->fxtype = TE_WIZSPIKE;
					p->drawtype = PFX_GLOW;
					p->color = 230 + ((tracercount&4)<<1);
					VectorCopy (start, p->org);
				}
				else if (type != 3 && fh1)
				{
					p->die = cl.time + 0.5;
					p->born = cl.time;
					p->fxtype = TE_KNIGHTSPIKE;
					p->drawtype = PFX_GLOW;
					p->color = 230 + ((tracercount&4)<<1);
					VectorCopy (start, p->org);
					for (j=0 ; j<3 ; j++)
						p->org[j] = start[j] + (((rand()&20)-10));
				}
				else if (type != 3 && !fh1)
				{
					p->die = cl.time + 1;
					p->born = cl.time;
					p->fxtype = TE_KNIGHTSPIKE;
					p->drawtype = PFX_GLOW;
					p->color = 230 + ((tracercount&4)<<1);
					VectorCopy (start, p->org);
				}
				tracercount++;
				break;

			case 4:	// slight blood
				p->fxtype = TE_BLOODTRAIL;
				p->drawtype = PFX_SPRITE;
				p->tex_num = zsmoke_tex;
				p->die = cl.time + 1;
				p->born = cl.time;
				p->type = pt_grav;
				p->color = 67 + (rand()&3);
				for (j=0 ; j<3 ; j++)
				{
					p->org[j] = start[j] + (((rand()%6)-3) );
					p->vel[j] = (rand()%10)-5;
				}
				len -= (50);
				break;

			case 6:	// voor trail
				p->fxtype = TE_SPARK;
				p->drawtype = PFX_GLOW;
				p->born = cl.time;
				p->color = 9*16 + 8 + (rand()&3);
				p->type = pt_static;
				p->die = cl.time + 0.25;
				for (j=0 ; j<3 ; j++)
					p->org[j] = start[j] + (((rand()&3)-6));
				break;
		}
		VectorAdd (start, vec, start);
	}
}

void ZR_Shock1 (vec3_t org, int type)
{
	int			i, j, count;
	zparticle_t	*p;


	NEWZPARTICLE;
	
	p->die = cl.time + 1;
	p->drawtype = PFX_GLOW;
	if (type == 1)
		p->fxtype = TE_SHOCKWAVE;
	else if (type == 2)
	{
		p->fxtype = TE_BIGSHOCKWAVE;
		p->die = cl.time + 0.125;
	}
	else
		p->fxtype = TE_SPARKWAVE;
	p->color = 7 + (rand()&7);
	p->type = pt_slowgrav;
	p->born = cl.time;
	for (j=0 ; j<3 ; j++)
	{
		p->org[j] = org[j];
		p->vel[j] = 0;
	}
}


void ZR_LavaSplash (vec3_t org)
{
	int			i, j, k, count;
	zparticle_t	*p;
	float		vel;
	vec3_t		dir;

	count = 4;

	for (i=-count ; i<count ; i++)
		for (j=-count ; j<count ; j++)
			for (k=0 ; k<1 ; k++)
			{

				NEWZPARTICLE;

				p->die = cl.time + 1;
				p->born = cl.time;
				p->color = 224 + (rand()&7);
				p->type = pt_slowgrav;
				p->fxtype = TE_LAVASPLASH;
				p->drawtype = PFX_TWINKLE;
				
				dir[0] = j*32 + (rand()&7);
				dir[1] = i*32 + (rand()&7);
				dir[2] = (rand()&20);
	
				p->org[0] = org[0] + (dir[0] );
				p->org[1] = org[1] + (dir[1] );
				p->org[2] = org[2] + ((rand()&63) );
	
				VectorNormalize (dir);						
				vel = 150 + (rand()&63);
				VectorScale (dir, vel, p->vel);
			}

	for (i=-count ; i<count ; i++)
		for (j=-count ; j<count ; j++)
			for (k=0 ; k<1 ; k++)
			{
			
				NEWZPARTICLE;
		
				p->die = cl.time + 1;
				p->born = cl.time;
				p->color = 224 + (rand()&7);
				p->type = pt_slowgrav;
				p->fxtype = TE_LAVASPLASH;
				p->drawtype = PFX_TWINKLE;
				
				dir[0] = j*8 + (rand()&7);
				dir[1] = i*8 + (rand()&7);
				dir[2] = 512;
	
				p->org[0] = org[0] + (dir[0]);
				p->org[1] = org[1] + (dir[1] );
				p->org[2] = org[2] + ((rand()&63) );
	
				VectorNormalize (dir);						
				vel = 150 + (rand()&63);
				VectorScale (dir, vel, p->vel);
			}
}


void ZR_Fire (vec3_t org, float power)// power=0.1, its ugly :(
{
	int			i;
	vec3_t		vel;
	zparticle_t	*p;
	
	for (i=0 ; i<power ; i++)
	{
		NEWZPARTICLE;

		p->die = cl.time + 0.5;
		p->born = cl.time;
		p->color = 224 + (rand()&7);
		p->type = pt_slowgrav;
		p->fxtype = TE_FIRE;
		p->drawtype = PFX_SPRITE;
		p->tex_num = zfire_tex + (rand()%6);
		p->angle = 0;
		p->v_angle = 0;
		
		p->org[0] = org[0] + ((((rand()%32) - 16) * power) );
		p->org[1] = org[1] + ((((rand()%32) - 16) * power) );
		p->org[2] = org[2] + (((rand()&16) - 16) );

		p->vel[0] = ((rand()%32) - 16);
		p->vel[1] = ((rand()%32) - 16);
		p->vel[2] = (rand()%16 + 8) * (power * 32);
	}
}


void ZR_Radar (vec3_t org, vec3_t dir) //????
{
	R_CreateParticle(1, TE_RADARBLIP, PFX_HUD, zradar_tex, 224 + (rand()&7), pt_static, cl.time, cl.time + 10, 0, 1.57, 1, 1, org, 0, 0, 0, 0);
}


void ZR_Blood (vec3_t org)
{
	int			i;
	vec3_t		vel;
	zparticle_t	*p;
	
  
		NEWZPARTICLE;

		p->die = cl.time + 12.5;
		p->born = cl.time;
		p->color = 224 + (rand()&7);
		p->type = pt_static;
		p->fxtype = 666;
		p->drawtype = PFX_SPRITE;
		p->tex_num =  zblood_tex  + (rand()%6);
		p->angle = 0;
		p->v_angle = 0;

		
		
		p->org[0] = org[0] + ((((rand()%32) - 16) ) );
		p->org[1] = org[1] + ((((rand()%32) - 16)) );
		p->org[2] = org[2] + (((rand()&16) - 16) );

		p->vel[0] = ((rand()%32) - 16);
		p->vel[1] = ((rand()%32) - 16);
		p->vel[2] = 0;
	
}
